#!/usr/bin/env python
# -*- coding: utf-8 -*-
import numpy as np
from sklearn.naive_bayes import BernoulliNB
from molalkit.models.base import BaseSklearnModel


class NBClassifier(BernoulliNB, BaseSklearnModel):
    def fit_alb(self, train_data):
        return self.fit_alb_(train_data, self)

    def predict_uncertainty(self, pred_data):
        return self.predict_uncertainty_c(pred_data, self)

    def predict_value(self, pred_data):
        return self.predict_value_c(pred_data, self)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os.path
import time
from molalkit.args import ActiveLearningArgs, ActiveLearningContinueArgs, ReEvaluateArgs
from molalkit.al.learner import ActiveLearner


def molalkit_run():
    run(args=ActiveLearningArgs().parse_args())


def run(args: ActiveLearningArgs):
    # active learning
    args.logger.info('Start a new active learning run.')
    start = time.time()
    if args.load_checkpoint and os.path.exists('%s/al.pkl' % args.save_dir):
        args.logger.info('continue active learning using checkpoint file %s/al.pkl' % args.save_dir)
        active_learner = ActiveLearner.load(path=args.save_dir)
    else:
        active_learner = ActiveLearner(save_dir=args.save_dir,
                                       selection_method=args.selection_method,
                                       forgetter=args.forgetter,
                                       model_selector=args.model_selector,
                                       dataset_train_selector=args.data_train_selector,
                                       dataset_pool_selector=args.data_pool_selector,
                                       dataset_val_selector=args.data_val_selector,
                                       metrics=args.metrics,
                                       top_k_id=args.top_k_id,
                                       model_evaluators=args.model_evaluators,
                                       dataset_train_evaluators=args.data_train_evaluators,
                                       dataset_pool_evaluators=args.data_pool_evaluators,
                                       dataset_val_evaluators=args.data_val_evaluators,
                                       yoked_learning_only=args.yoked_learning_only,
                                       stop_size=args.stop_size,
                                       evaluate_stride=args.evaluate_stride,
                                       kernel=args.kernel_selector,
                                       save_cpt_stride=args.save_cpt_stride,
                                       seed=args.seed,
                                       logger=args.logger)
    active_learner.run(max_iter=args.max_iter)
    end = time.time()
    args.logger.info('total time: %d s' % (end - start))
    return active_learner


def run_from_cpt():
    # read args.
    args = ActiveLearningContinueArgs().parse_args()
    # active learning
    args.logger.info('Start a new active learning run.')
    start = time.time()
    args.logger.info('continue active learning using checkpoint file %s/al.pkl' % args.save_dir)
    active_learner = ActiveLearner.load(path=args.save_dir)
    # change stop_size.
    if args.stop_ratio is not None:
        if args.stop_size is None:
            args.stop_size = int(args.stop_ratio * (active_learner.train_size + active_learner.pool_size))
        else:
            args.stop_size = min(args.stop_size,
                                 int(args.stop_ratio * (active_learner.train_size + active_learner.pool_size)))
        assert args.stop_size >= 2
    active_learner.stop_size = args.stop_size
    active_learner.run()
    end = time.time()
    args.logger.info('total time: %d s' % (end - start))


def run_eval():
    pass
    # read args.
    """
    args = ReEvaluateArgs().parse_args()
    args.logger.info('Start a new active learning run.')
    data_train = args.data_al_evaluator.copy()
    active_learning_traj_dict = {'training_size': []}
    for metric in args.metrics:
        active_learning_traj_dict[metric] = []
    n_list = list(range(args.evaluate_stride, len(args.data_al_evaluator), args.evaluate_stride))
    if n_list[-1] != len(data_train):
        n_list.append(len(data_train))
    for n in tqdm(n_list):
        active_learning_traj_dict['training_size'].append(n)
        data_train.data = args.data_al_evaluator.data[:n]
        args.model_evaluator.fit(data_train)
        y_pred = args.model_evaluator.predict_value(args.data_val_evaluator)
        for metric in args.metrics:
            metric_value = eval_metric_func(args.data_val_evaluator.y, y_pred, metric=metric)
            active_learning_traj_dict[metric].append(metric_value)
    pd.DataFrame(active_learning_traj_dict).to_csv(
        '%s/active_learning_extra_%d.traj' % (args.save_dir, args.evaluator_id), index=False)
    """



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import pytest
import os
import shutil
from molalkit.args import ActiveLearningArgs
from molalkit.al.learner import ActiveLearner
from molalkit.data.datasets import DATA_DIR
from test_model.test_model import al_results_check


CWD = os.path.dirname(os.path.abspath(__file__))


@pytest.mark.parametrize('batch_size', ['1', '5', '20'])
def test_classification(batch_size):
    save_dir = os.path.join(CWD, 'test')
    arguments = [
        '--data_path_training', f'{DATA_DIR}/carcinogens_lagunin.csv',
        '--data_path_pool', f'{DATA_DIR}/skin.csv',
        '--pure_columns', 'Drug',
        '--target_columns', 'Y',
        '--metrics', 'roc-auc',
        '--learning_type', 'explorative',
        '--model_config_selector', 'RandomForest_RDKitNorm_Config',
        '--split_type', 'random',
        '--split_sizes', '1.0', '0.0',
        '--evaluate_stride', '1',
        '--batch_size', batch_size,
        '--stop_size', '300',
        '--seed', '0',
        '--save_dir', save_dir,
        '--n_jobs', '4'
    ]
    args = ActiveLearningArgs().parse_args(arguments)
    active_learner = ActiveLearner(save_dir=args.save_dir,
                                   selection_method=args.selection_method,
                                   forgetter=args.forgetter,
                                   model_selector=args.model_selector,
                                   dataset_train_selector=args.data_train_selector,
                                   dataset_pool_selector=args.data_pool_selector,
                                   dataset_val_selector=args.data_val_selector,
                                   metrics=args.metrics,
                                   top_k_id=args.top_k_id,
                                   model_evaluators=args.model_evaluators,
                                   dataset_train_evaluators=args.data_train_evaluators,
                                   dataset_pool_evaluators=args.data_pool_evaluators,
                                   dataset_val_evaluators=args.data_val_evaluators,
                                   yoked_learning_only=args.yoked_learning_only,
                                   stop_size=args.stop_size,
                                   evaluate_stride=args.evaluate_stride,
                                   kernel=args.kernel_selector,
                                   save_cpt_stride=args.save_cpt_stride,
                                   seed=args.seed,
                                   logger=args.logger)
    alr = active_learner.run_prospective()
    batch_size = int(batch_size)
    assert len(alr.id_add) == batch_size
    ids = [data.id for data in active_learner.dataset_train_selector[-batch_size:]]
    assert set(ids) == set(alr.id_add)
    shutil.rmtree(f'{save_dir}')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import pytest
import os
import shutil
from molalkit.args import ActiveLearningArgs
from test_model.test_model import run, al_results_check


CWD = os.path.dirname(os.path.abspath(__file__))


@pytest.mark.parametrize('split', ['random', 'scaffold_order', 'scaffold_random'])
def test_classification(split):
    save_dir = os.path.join(CWD, 'test')
    arguments = [
        '--data_public', 'carcinogens_lagunin',
        '--metrics', 'roc-auc',
        '--learning_type', 'explorative',
        '--model_config_selector', 'RandomForest_RDKitNorm_Config',
        '--split_type', split,
        '--split_sizes', '0.5', '0.5',
        '--evaluate_stride', '1',
        '--stop_size', '5',
        '--seed', '0',
        '--save_dir', save_dir,
        '--n_jobs', '4'
    ]
    args = ActiveLearningArgs().parse_args(arguments)
    active_learner = run(args)
    assert len(active_learner.active_learning_traj.results) == 3
    al_results_check(save_dir)
    shutil.rmtree(f'{save_dir}')


def test_classification_full():
    save_dir = os.path.join(CWD, 'test')
    arguments = [
        '--data_public', 'carcinogens_lagunin',
        '--metrics', 'roc-auc',
        '--learning_type', 'explorative',
        '--model_config_selector', 'RandomForest_RDKitNorm_Config',
        '--full_val',
        '--evaluate_stride', '1',
        '--stop_size', '5',
        '--seed', '0',
        '--save_dir', save_dir,
        '--n_jobs', '4'
    ]
    args = ActiveLearningArgs().parse_args(arguments)
    active_learner = run(args)
    assert (len(active_learner.dataset_train_selector) + len(active_learner.dataset_pool_selector) ==
            len(active_learner.dataset_val_selector))
    assert len(active_learner.active_learning_traj.results) == 3
    al_results_check(save_dir)
    shutil.rmtree(f'{save_dir}')


def test_classification_no_val():
    save_dir = os.path.join(CWD, 'test')
    arguments = [
        '--data_public', 'carcinogens_lagunin',
        '--learning_type', 'explorative',
        '--model_config_selector', 'RandomForest_RDKitNorm_Config',
        '--split_type', 'random',
        '--split_sizes', '1.0', '0.0',
        '--stop_size', '5',
        '--seed', '0',
        '--save_dir', save_dir,
        '--n_jobs', '4'
    ]
    args = ActiveLearningArgs().parse_args(arguments)
    assert args.data_val_selector is None
    active_learner = run(args)
    assert active_learner.dataset_val_selector is None
    assert len(active_learner.active_learning_traj.results) == 3
    al_results_check(save_dir)
    shutil.rmtree(f'{save_dir}')


@pytest.mark.parametrize('split', ['random', 'scaffold_order', 'scaffold_random'])
def test_regression(split):
    save_dir = os.path.join(CWD, 'test')
    arguments = [
        '--data_public', 'test_regression',
        '--metrics', 'rmse',
        '--learning_type', 'explorative',
        '--model_config_selector', 'RandomForest_RDKitNorm_Config',
        '--split_type', split,
        '--split_sizes', '0.5', '0.5',
        '--evaluate_stride', '1',
        '--stop_size', '5',
        '--seed', '0',
        '--save_dir', save_dir,
        '--n_jobs', '4'
    ]
    args = ActiveLearningArgs().parse_args(arguments)
    active_learner = run(args)
    assert len(active_learner.active_learning_traj.results) == 3
    al_results_check(save_dir)
    shutil.rmtree(f'{save_dir}')


def test_regression_full():
    save_dir = os.path.join(CWD, 'test')
    arguments = [
        '--data_public', 'test_regression',
        '--metrics', 'rmse',
        '--learning_type', 'explorative',
        '--model_config_selector', 'RandomForest_RDKitNorm_Config',
        '--full_val',
        '--evaluate_stride', '1',
        '--stop_size', '5',
        '--seed', '0',
        '--save_dir', save_dir,
        '--n_jobs', '4'
    ]
    args = ActiveLearningArgs().parse_args(arguments)
    active_learner = run(args)
    assert (len(active_learner.dataset_train_selector) + len(active_learner.dataset_pool_selector) ==
            len(active_learner.dataset_val_selector))
    assert len(active_learner.active_learning_traj.results) == 3
    al_results_check(save_dir)
    shutil.rmtree(f'{save_dir}')


def test_regression_no_val():
    save_dir = os.path.join(CWD, 'test')
    arguments = [
        '--data_public', 'test_regression',
        '--learning_type', 'explorative',
        '--model_config_selector', 'RandomForest_RDKitNorm_Config',
        '--split_type', 'random',
        '--split_sizes', '1.0', '0.0',
        '--stop_size', '5',
        '--seed', '0',
        '--save_dir', save_dir,
        '--n_jobs', '4'
    ]
    args = ActiveLearningArgs().parse_args(arguments)
    assert args.data_val_selector is None
    active_learner = run(args)
    assert active_learner.dataset_val_selector is None
    assert len(active_learner.active_learning_traj.results) == 3
    al_results_check(save_dir)
    shutil.rmtree(f'{save_dir}')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import glob


MODEL_DIR = os.path.dirname(__file__)
file_list = glob.glob(os.path.join(MODEL_DIR, "*Config"))
AVAILABLE_MODELS = sorted([os.path.basename(file) for file in file_list])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from abc import ABC, abstractmethod
from typing import Dict, Iterator, List, Optional, Union, Literal, Tuple, Callable
import numpy as np
from collections import defaultdict
from sklearn.cluster import KMeans
from sklearn.manifold import SpectralEmbedding


def get_topn_idx(values: np.ndarray, n: int = 1, target: Union[Literal['max', 'min'], float] = 'max',
                 cutoff: float = None) -> List[int]:
    if isinstance(values, list):
        values = np.array(values)
    if target == 'min':
        values = - values
        if cutoff is not None:
            cutoff = - cutoff
    elif isinstance(target, float):
        assert cutoff is None
        values = np.absolute(values - target)  # distance from target
    if cutoff is not None:
        n_candidates = len(np.where(values >= cutoff)[0])
        n = min(n, n_candidates)
        if n == 0:
            return []
    # Includes tiny random values to randomly sort duplicated values
    sorting_key = values + np.random.random(len(values)) * 1e-10
    return np.argsort(sorting_key)[-n:].tolist()


class BaseSelectionMethod(ABC):
    @abstractmethod
    def __call__(self, **kwargs):
        pass

    @property
    @abstractmethod
    def info(self) -> str:
        pass


class BaseRandomSelectionMethod(BaseSelectionMethod, ABC):
    """Base SelectionMethod that uses random seed."""
    def __init__(self, batch_size: int = 1, seed: int = 0):
        self.batch_size = batch_size
        np.random.seed(seed)


class BaseClusterSelectionMethod(BaseRandomSelectionMethod, ABC):
    def __init__(self, batch_size: int = 1, cluster_size: int = None, seed: int = 0):
        super().__init__(batch_size=batch_size, seed=seed)
        if cluster_size is None:
            self.cluster_size = batch_size * 20
        else:
            self.cluster_size = cluster_size

    @staticmethod
    def find_distant_samples(gram_matrix: List[List[float]], batch_size: int = 1) -> List[int]:
        """ Find distant samples from a pool using clustering method.

        Parameters
        ----------
        gram_matrix: gram (kernel) matrix of the samples.
        batch_size: number of samples to be selected.

        Returns
        -------
        List of idx
        """
        embedding = SpectralEmbedding(
            n_components=batch_size,
            affinity='precomputed'
        ).fit_transform(gram_matrix)

        cluster_result = KMeans(
            n_clusters=batch_size,
            # random_state=self.args.seed
        ).fit_predict(embedding)
        # find all center of clustering
        center = np.array([embedding[cluster_result == i].mean(axis=0)
                           for i in range(batch_size)])
        total_distance = defaultdict(dict)  # (key: cluster_idx, val: dict of (key:sum of distance, val:idx))
        for i in range(len(cluster_result)):
            cluster_class = cluster_result[i]
            total_distance[cluster_class][((np.square(
                embedding[i] - np.delete(center, cluster_class, axis=0))).sum(
                axis=1) ** -0.5).sum()] = i
        add_idx = [total_distance[i][min(total_distance[i].keys())] for i in
                   range(batch_size)]  # find min-in-cluster-distance associated idx
        return add_idx


class BaseIterativeSelectionMethod(BaseSelectionMethod, ABC):
    pass


class RandomSelectionMethod(BaseRandomSelectionMethod):
    def __call__(self, data_pool, **kwargs) -> Tuple[List[int], None]:
        if self.batch_size < len(data_pool):
            return np.random.choice(range(len(data_pool)), self.batch_size, replace=False).tolist(), None
        else:
            return list(range(len(data_pool))), None

    @property
    def info(self) -> str:
        return f'RandomSelectionMethod(batch_size={self.batch_size})'


class ClusterRandomSelectionMethod(BaseClusterSelectionMethod):
    def __call__(self, data_pool, kernel: Callable) -> Tuple[List[int], None]:
        assert self.batch_size < len(data_pool)
        idx_candidates = np.random.choice(range(len(data_pool)), self.cluster_size, replace=False).tolist()
        K = kernel(data_pool.X[idx_candidates])
        add_idx = self.find_distant_samples(gram_matrix=K, batch_size=self.batch_size)
        idx = np.array(idx_candidates)[add_idx]
        return idx, None

    @property
    def info(self) -> str:
        return f'ClusterRandomSelectionMethod(batch_size={self.batch_size}, cluster_size={self.cluster_size})'


class ExplorativeSelectionMethod(BaseRandomSelectionMethod):
    def __call__(self, model, data_pool, **kwargs) -> Tuple[List[int], List[float]]:
        y_std = model.predict_uncertainty(data_pool)
        idx = get_topn_idx(y_std, n=self.batch_size)
        acquisition = y_std[np.array(idx)].tolist()
        return idx, acquisition

    @property
    def info(self) -> str:
        return f'ExplorativeSelectionMethod(batch_size={self.batch_size})'


class ClusterExplorativeSelectionMethod(BaseClusterSelectionMethod):
    def __call__(self, model, data_pool, kernel: Callable, **kwargs) -> Tuple[List[int], List[float]]:
        y_std = model.predict_uncertainty(data_pool)
        idx_candidates = get_topn_idx(y_std, n=self.cluster_size)
        K = kernel(data_pool.X[idx_candidates])
        add_idx = self.find_distant_samples(gram_matrix=K, batch_size=self.batch_size)
        idx = np.array(idx_candidates)[add_idx]
        acquisition = y_std[idx].tolist()
        return idx, acquisition

    @property
    def info(self) -> str:
        return f'ClusterExplorativeSelectionMethod(batch_size={self.batch_size}, cluster_size={self.cluster_size})'


class ExploitiveSelectionMethod(BaseRandomSelectionMethod):
    def __init__(self, target, seed: int = 0):
        super().__init__(seed=seed)
        self.target = target

    def __call__(self, model, data_pool, **kwargs) -> Tuple[List[int], List[float]]:
        y_pred = model.predict_value(data_pool)
        idx = get_topn_idx(y_pred, n=self.batch_size, target=self.target)
        acquisition = y_pred[np.array(idx)].tolist()
        return idx, acquisition

    @property
    def info(self) -> str:
        return f'ExploitiveSelectionMethod(batch_size={self.batch_size}, target={self.target})'


class ClusterExploitiveSelectionMethod(BaseClusterSelectionMethod):
    def __init__(self, target, seed: int = 0):
        super().__init__(seed=seed)
        self.target = target

    def __call__(self, model, data_pool, kernel: Callable, **kwargs) -> Tuple[List[int], List[float]]:
        y_pred = model.predict_value(data_pool)
        idx_candidates = get_topn_idx(y_pred, n=self.cluster_size, target=self.target)
        K = kernel(data_pool.X[idx_candidates])
        add_idx = self.find_distant_samples(gram_matrix=K, batch_size=self.batch_size)
        idx = np.array(idx_candidates)[add_idx]
        acquisition = y_pred[idx].tolist()
        return idx, acquisition

    @property
    def info(self) -> str:
        return (f'ClusterExploitiveSelectionMethod(batch_size={self.batch_size}, '
                f'cluster_size={self.cluster_size}, target={self.target})')


class ProbabilityImprovementSelectionMethod(BaseRandomSelectionMethod):
    pass


class ExpectedImprovementSelectionMethod(BaseRandomSelectionMethod):
    pass


class UpperConfidenceBoundSelectionMethod(BaseRandomSelectionMethod):
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import glob


DATA_DIR = os.path.dirname(__file__)
file_list = glob.glob(os.path.join(DATA_DIR, "*.csv"))
AVAILABLE_DATASETS = sorted([os.path.basename(file) for file in file_list])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import pytest
import os
import shutil
import pandas as pd
from molalkit.args import ActiveLearningArgs
from test_model.test_model import run, al_results_check


CWD = os.path.dirname(os.path.abspath(__file__))


@pytest.mark.parametrize('error_rate', ['0.1', '0.2', '0.3'])
def test_classification(error_rate):
    save_dir = os.path.join(CWD, 'test')
    arguments = [
        '--data_public', 'carcinogens_lagunin',
        '--metrics', 'roc-auc',
        '--learning_type', 'explorative',
        '--model_config_selector', 'RandomForest_RDKitNorm_Config',
        '--split_type', 'random',
        '--split_sizes', '0.5', '0.5',
        '--error_rate', error_rate,
        '--evaluate_stride', '1',
        '--stop_size', '5',
        '--seed', '0',
        '--save_dir', save_dir,
        '--n_jobs', '4'
    ]
    args = ActiveLearningArgs().parse_args(arguments)
    active_learner = run(args)
    assert len(active_learner.active_learning_traj.results) == 3
    al_results_check(save_dir)
    df1 = pd.read_csv('%s/train_init.csv' % save_dir)
    df2 = pd.read_csv('%s/pool_init.csv' % save_dir)
    df = pd.concat([df1, df2])
    df_ = pd.read_csv('%s/val.csv' % save_dir)
    assert abs(len(df) - len(df_)) <= 1
    assert df['flip_label'].sum() == int(float(error_rate) * len(df))
    shutil.rmtree(f'{save_dir}')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from typing import Callable, List, Union

import numpy as np
from rdkit import Chem, DataStructs
from rdkit.Chem import AllChem
from descriptastorus.descriptors import rdDescriptors, rdNormalizedDescriptors


def morgan_binary_features_generator(mol: Union[str, Chem.Mol],
                                     radius: int,
                                     num_bits: int) -> np.ndarray:
    """
    Generates a binary Morgan fingerprint for a molecule.

    :param mol: A molecule (i.e., either a SMILES or an RDKit molecule).
    :param radius: Morgan fingerprint radius.
    :param num_bits: Number of bits in Morgan fingerprint.
    :return: A 1D numpy array containing the binary Morgan fingerprint.
    """
    mol = Chem.MolFromSmiles(mol) if type(mol) == str else mol
    features_vec = AllChem.GetMorganFingerprintAsBitVect(mol, radius, nBits=num_bits)
    features = np.zeros((1,))
    DataStructs.ConvertToNumpyArray(features_vec, features)

    return features


def morgan_counts_features_generator(mol: Union[str, Chem.Mol],
                                     radius: int,
                                     num_bits: int) -> np.ndarray:
    """
    Generates a counts-based Morgan fingerprint for a molecule.

    :param mol: A molecule (i.e., either a SMILES or an RDKit molecule).
    :param radius: Morgan fingerprint radius.
    :param num_bits: Number of bits in Morgan fingerprint.
    :return: A 1D numpy array containing the counts-based Morgan fingerprint.
    """
    mol = Chem.MolFromSmiles(mol) if type(mol) == str else mol
    features_vec = AllChem.GetHashedMorganFingerprint(mol, radius, nBits=num_bits)
    features = np.zeros((1,))
    DataStructs.ConvertToNumpyArray(features_vec, features)

    return features


def rdkit_2d_features_generator(mol: Union[str, Chem.Mol]) -> np.ndarray:
    """
    Generates RDKit 2D features_mol for a molecule.

    :param mol: A molecule (i.e., either a SMILES or an RDKit molecule).
    :return: A 1D numpy array containing the RDKit 2D features_mol.
    """
    smiles = Chem.MolToSmiles(mol, isomericSmiles=True) if type(mol) != str else mol
    generator = rdDescriptors.RDKit2D()
    features = generator.process(smiles)[1:]

    return features


def rdkit_2d_normalized_features_generator(mol: Union[str, Chem.Mol]) -> np.ndarray:
    """
    Generates RDKit 2D normalized features_mol for a molecule.

    :param mol: A molecule (i.e., either a SMILES or an RDKit molecule).
    :return: A 1D numpy array containing the RDKit 2D normalized features_mol.
    """
    smiles = Chem.MolToSmiles(mol, isomericSmiles=True) if type(mol) != str else mol
    generator = rdNormalizedDescriptors.RDKit2DNormalized()
    features = generator.process(smiles)[1:]

    return features



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import time
from typing import Dict, Iterator, List, Optional, Union, Literal, Tuple
from tqdm import trange
from logging import Logger
import numpy as np
import torch
from tensorboardX import SummaryWriter
from torch.optim.lr_scheduler import ExponentialLR
from chemprop.data import get_class_sizes, get_data, MoleculeDataLoader, MoleculeDataset, set_cache_graph, split_data, \
    get_task_names
from chemprop.utils import build_optimizer, build_lr_scheduler, load_checkpoint, makedirs, \
    save_checkpoint, save_smiles_splits, load_frzn_model, multitask_mean
from chemprop.nn_utils import param_count, param_count_all
from chemprop.models import MoleculeModel
from chemprop.constants import MODEL_FILE_NAME
from chemprop.train.loss_functions import get_loss_func
from chemprop.train import train
from chemprop.train.predict import predict
from .args import TrainArgs


class MPNN:
    def __init__(self,
                 save_dir: str,
                 dataset_type: Literal['regression', 'classification', 'multiclass', 'spectra'],
                 loss_function: Literal['mse', 'bounded_mse', 'binary_cross_entropy', 'cross_entropy', 'mcc', 'sid',
                                        'wasserstein', 'mve', 'evidential', 'dirichlet'],
                 num_tasks: int = 1,
                 multiclass_num_classes: int = 3,
                 features_generator=None,
                 no_features_scaling: bool = False,
                 features_only: bool = False,
                 features_size: int = 0,
                 epochs: int = 30,
                 depth: int = 3,
                 hidden_size: int = 300,
                 ffn_num_layers: int = 2,
                 ffn_hidden_size: int = None,
                 dropout: float = 0.0,
                 batch_size: int = 50,
                 ensemble_size: int = 1,
                 number_of_molecules: int = 1,
                 mpn_shared: bool = False,
                 atom_messages: bool = False,
                 undirected: bool = False,
                 n_jobs: int = 8,
                 class_balance: bool = False,
                 checkpoint_dir: str = None,
                 checkpoint_frzn: str = None,
                 frzn_ffn_layers: int = 0,
                 freeze_first_only: bool = False,
                 seed: int = 0,
                 logger: Logger = None,
                 ):
        """
        args = TrainArgs(save_dir=save_dir,
                         dataset_type=dataset_type,
                         loss_function=loss_function,
                         multiclass_num_classes=multiclass_num_classes,
                         features_only=features_only,
                         epochs=epochs,
                         hidden_size=hidden_size,
                         ffn_num_layers=ffn_num_layers,
                         ffn_hidden_size=ffn_hidden_size,
                         dropout=dropout,
                         batch_size=batch_size,
                         ensemble_size=ensemble_size,
                         number_of_molecules=number_of_molecules,
                         mpn_shared=mpn_shared,
                         atom_messages=atom_messages,
                         undirected=undirected,
                         num_workers=n_jobs,
                         class_balance=class_balance,
                         checkpoint_dir=checkpoint_dir,
                         checkpoint_frzn=checkpoint_frzn,
                         frzn_ffn_layers=frzn_ffn_layers,
                         freeze_first_only=freeze_first_only,
                         seed=seed)
        """
        args = TrainArgs()
        args.save_dir = save_dir
        args.dataset_type = dataset_type
        args.loss_function = loss_function
        args.num_tasks = num_tasks
        args.multiclass_num_classes = multiclass_num_classes
        args.features_generator = features_generator
        args.no_features_scaling = no_features_scaling
        args.features_only = features_only
        args.features_size = features_size
        args.epochs = epochs
        args.depth = depth
        args.hidden_size = hidden_size
        args.ffn_num_layers = ffn_num_layers
        args.ffn_hidden_size = ffn_hidden_size
        args.dropout = dropout
        args.batch_size = batch_size
        args.ensemble_size = ensemble_size
        args.number_of_molecules = number_of_molecules
        args.mpn_shared = mpn_shared
        args.atom_messages = atom_messages
        args.undirected = undirected
        args.num_workers = n_jobs
        args.class_balance = class_balance
        args.checkpoint_dir = checkpoint_dir
        args.checkpoint_frzn = checkpoint_frzn
        args.frzn_ffn_layers = frzn_ffn_layers
        args.freeze_first_only = freeze_first_only
        args.seed = seed
        args.process_args()
        self.args = args
        self.features_scaler = None
        self.logger = logger

    def fit_alb(self, train_data):
        if torch.cuda.is_available():
            torch.cuda.empty_cache()
        args = self.args
        args.train_data_size = len(train_data)
        logger = self.logger
        if logger is not None:
            debug, info = logger.debug, logger.info
        else:
            debug = info = str

        # Set pytorch seed for random initial weights
        torch.manual_seed(args.pytorch_seed)

        if args.dataset_type == 'classification':
            train_class_sizes = get_class_sizes(train_data, proportion=False)
            args.train_class_sizes = train_class_sizes

        if args.features_scaling:
            self.features_scaler = train_data.normalize_features(replace_nan_token=0)

        atom_descriptor_scaler = None
        bond_feature_scaler = None

        args.train_data_size = len(train_data)

        # Initialize scaler and scale training targets by subtracting mean and dividing standard deviation (
        # regression only)
        if args.dataset_type == 'regression':
            debug('Fitting scaler')
            scaler = train_data.normalize_targets()
            args.spectra_phase_mask = None
        else:
            args.spectra_phase_mask = None
            scaler = None

        # Get loss function
        loss_func = get_loss_func(args)
        """
        # Automatically determine whether to cache
        if len(train_data) <= args.cache_cutoff:
            set_cache_graph(True)
            num_workers = 0
        else:
            set_cache_graph(False)
            num_workers = args.num_workers
        """
        train_data_loader = MoleculeDataLoader(
            dataset=train_data,
            batch_size=args.batch_size,
            num_workers=args.num_workers,
            class_balance=args.class_balance,
            shuffle=True,
            seed=args.seed
        )

        if args.class_balance:
            debug(f'With class_balance, effective train size = {train_data_loader.iter_size:,}')

        for model_idx in range(args.ensemble_size):
            # Tensorboard writer
            save_dir = os.path.join(args.save_dir, f'model_{model_idx}')
            makedirs(save_dir)
            #try:
            #    writer = SummaryWriter(log_dir=save_dir)
            #except:
            #    writer = SummaryWriter(logdir=save_dir)
            writer = None
            # Load/build model
            if args.checkpoint_paths is not None:
                debug(f'Loading model {model_idx} from {args.checkpoint_paths[model_idx]}')
                model = load_checkpoint(args.checkpoint_paths[model_idx], logger=logger)
            else:
                debug(f'Building model {model_idx}')
                model = MoleculeModel(args)

            # Optionally, overwrite weights:
            if args.checkpoint_frzn is not None:
                debug(f'Loading and freezing parameters from {args.checkpoint_frzn}.')
                model = load_frzn_model(model=model, path=args.checkpoint_frzn, current_args=args, logger=logger)

            debug(model)

            if args.checkpoint_frzn is not None:
                debug(f'Number of unfrozen parameters = {param_count(model):,}')
                debug(f'Total number of parameters = {param_count_all(model):,}')
            else:
                debug(f'Number of parameters = {param_count_all(model):,}')

            if args.cuda:
                print('Moving model to cuda')
            model = model.to(args.device)

            # Ensure that model is saved in correct location for evaluation if 0 epochs
            # save_checkpoint(os.path.join(save_dir, MODEL_FILE_NAME), model, scaler,
            #                 features_scaler, atom_descriptor_scaler, bond_feature_scaler, None)

            # Optimizers
            optimizer = build_optimizer(model, args)

            # Learning rate schedulers
            scheduler = build_lr_scheduler(optimizer, args)

            n_iter = 0
            for epoch in trange(args.epochs):
                debug(f'Epoch {epoch}')
                n_iter = train(
                    model=model,
                    data_loader=train_data_loader,
                    loss_func=loss_func,
                    optimizer=optimizer,
                    scheduler=scheduler,
                    args=args,
                    n_iter=n_iter,
                    logger=logger,
                    writer=writer
                )
                if isinstance(scheduler, ExponentialLR):
                    scheduler.step()
            self.model = model
            self.scaler = scaler
            # save_checkpoint(os.path.join(save_dir, MODEL_FILE_NAME), model, scaler, features_scaler,
            #                 atom_descriptor_scaler, bond_feature_scaler, None)

    def predict_uncertainty(self, pred_data):
        args = self.args
        if args.features_scaling:
            pred_data.normalize_features(self.features_scaler)
        pred_data_loader = MoleculeDataLoader(
            dataset=pred_data,
            batch_size=args.batch_size,
            num_workers=args.num_workers
        )
        preds = predict(
            model=self.model,
            data_loader=pred_data_loader,
            scaler=self.scaler,
            return_unc_parameters=True
        )
        if self.args.dataset_type == 'classification':
            preds = np.asarray(preds)
            preds = np.concatenate([preds, 1-preds], axis=1)
            return 0.25 - np.var(preds, axis=1)
        elif self.args.dataset_type == 'multiclass':
            return 0.25 - np.var(preds, axis=1)
        elif self.args.dataset_type == 'regression':
            if self.model.loss_function == "mve":
                preds, var = preds
                return np.array(var).ravel()
            elif self.args.loss_function == 'evidential':
                preds, lambdas, alphas, betas = preds
                return (np.array(betas) / (np.array(lambdas) * (np.array(alphas) - 1))).ravel()
            else:
                raise ValueError
        else:
            raise ValueError

    def predict_value(self, pred_data):
        args = self.args
        if args.features_scaling:
            pred_data.normalize_features(self.features_scaler)
        pred_data_loader = MoleculeDataLoader(
            dataset=pred_data,
            batch_size=args.batch_size,
            num_workers=args.num_workers
        )
        preds = predict(
            model=self.model,
            data_loader=pred_data_loader,
            scaler=self.scaler
        )
        if self.args.dataset_type in ['classification', 'multiclass']:
            return np.asarray(preds).ravel()
        elif self.args.dataset_type == 'regression':
            return np.asarray(preds).ravel()
        else:
            raise ValueError()



import io
import os
import re
import setuptools

with open('molalkit/__init__.py') as fd:
    __version__ = re.search("__version__ = '(.*)'", fd.read()).group(1)


def read(*filenames, **kwargs):
    encoding = kwargs.get('encoding', 'utf-8')
    sep = kwargs.get('sep', '\n')
    buf = []
    for filename in filenames:
        with io.open(filename, encoding=encoding) as f:
            buf.append(f.read())
    return sep.join(buf)


long_description = read('README.md')

setuptools.setup(
    name='molalkit',
    version=__version__,
    python_requires='>=3.8',
    install_requires=[
        'typed-argument-parser',
        'rdkit',
    ],
    entry_points={
        'console_scripts': [
            'molalkit_run=molalkit.al.run:molalkit_run',
        ]
    },
    author='Yan Xiang',
    author_email='yan.xiang@duke.edu',
    description='MolALKit: A Toolkit for Active Learning in Molecular Data.',
    long_description=long_description,
    url='https://github.com/RekerLab/MolALKit',
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    include_package_data=True,
    package_data={'': ['data/datasets/*.csv', 'models/configs/*Config']}
)
